
// gcc -Wall -Wextra -O3 mutex_test.c -o mutex_test

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include <sys/time.h>

#ifdef _WIN32
#  include <windows.h>
  typedef HANDLE generic_mtx;
#else
#  include <pthread.h>
  typedef pthread_mutex_t generic_mtx;
#endif /* _WIN32 */

#define BUILD_64_BIT 64

typedef int64_t cpu_ticks;

static float f_clock = 2.9; // 2.9 GHz

/* Pentium cycle counter */
static __inline__ cpu_ticks get_cpu_ticks(void)
{
    cpu_ticks ret;
#  if defined(BUILD_64_BIT)
    uint32_t _hi, _lo;
    __asm__ __volatile__("rdtsc": "=a" (_lo), "=d" (_hi));
    ret = ((cpu_ticks)_hi << 32) | _lo;
#  else // 32-bit intel version
    __asm__ __volatile__("rdtsc": "=A" (ret));
    /* no input, nothing else clobbered */
#  endif
    return ret;
}

generic_mtx *create_mtx(void)
{
    /* Allocate storage */
    generic_mtx *px_newres = (generic_mtx *)malloc(sizeof(generic_mtx));
#ifdef _WIN32
    if (px_newres != NULL)
    {
        /* Populate the newly allocated memory */
        *px_newres = CreateMutex (NULL, FALSE, NULL);
    }
    /* Return pointer to the new item, or NULL if creation failed */
    return px_newres;
#else
    int res_mtx = -1;               /* Error codes */
    if (px_newres != NULL)
    {
        res_mtx = pthread_mutex_init (px_newres, NULL); /* use defaults... */
        if(res_mtx == 0)
        {
            return px_newres;
        }
        else
        {
            /* Mutex creation failed */
            return NULL;
        }
    }
    else
    {
        /* Memory allocation failed */
        return NULL;
    }
#endif /* _WIN32 */
} // create_mtx

int lock_mtx (generic_mtx *px_res)
{
    int t_status = (-1);
#ifdef _WIN32
    DWORD i_result;                          /* Error code returned by function */
    /* Check Parameter */
    if ((px_res != NULL) && (*px_res != NULL))
    {
        /* Attempt to take the mutex with no timeout */
        i_result = WaitForSingleObject(*px_res, 0);
        if ((i_result == WAIT_OBJECT_0) || (i_result == WAIT_ABANDONED))
        {
            t_status = 0;
        }
    }
#else
    int i_result;
    if (px_res != NULL)
    {
        /* blocking lock */
        i_result = pthread_mutex_lock(px_res);
        if (i_result == 0)
        {
            t_status = 0;
        }
    }
#endif /* _WIN32 */
    return t_status;
} // lock_mtx

int unlock_mtx (generic_mtx *px_res)
{
    int t_status = (-1);
#ifdef _WIN32
    if (px_res != NULL)
    {
        /* Release the resource */
        if (ReleaseMutex(*px_res) != 0)
        {
            /* If the resource has been released, return 0 */
            t_status = 0;
        }
    }
#else
    int i_result;
    if (px_res != NULL)
    {
        i_result = pthread_mutex_unlock(px_res);
        if (i_result == 0)
        {
            t_status = 0;
        }
    }
#endif /* _WIN32 */
    return t_status;
} /* unlock_mtx */


static void do_loop_iters (const int max_iters, generic_mtx *p_mutex)
{
  cpu_ticks lock_time = 0;
  cpu_ticks unlock_time = 0;
  for (int i = 0; i < max_iters; ++i)
  {
      cpu_ticks before_lock = get_cpu_ticks();
      lock_mtx (p_mutex);
      cpu_ticks after_lock = get_cpu_ticks();
      lock_time += after_lock - before_lock;

      before_lock = get_cpu_ticks();
      unlock_mtx (p_mutex);
      after_lock = get_cpu_ticks();
      unlock_time += after_lock - before_lock;
  }
  long lt = (long)(lock_time / max_iters);
  long ut = (long)(unlock_time / max_iters);

  float lf = (float)lt / f_clock;
  float uf = (float)ut / f_clock;

  printf("Lock time   : %ld (%d iterations) -- %.2fns\n", lt, max_iters, lf);
  printf("Unlock time : %ld (%d iterations) -- %.2fns\n", ut, max_iters, uf);
} // do_loop_iters

int main (int argc, char **argv)
{
  if (argc > 1)
  {
    float speed = atof (argv[1]);
    f_clock = speed;
  }

  generic_mtx *p_mutex = create_mtx();

  do_loop_iters (1, p_mutex);
  do_loop_iters (10, p_mutex);
  do_loop_iters (100, p_mutex);

  return (0);
} // main

/* End of file */
